#include <stdio.h>
#include <pthread.h>
#include <stdint.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdatomic.h>

#define THREADS_COUNT 2

atomic_int processed_count = 0;
pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER; 
pthread_cond_t cond = PTHREAD_COND_INITIALIZER; 

void Process() {
  for (;;) {
    printf("Processor: Waiting for a signal...\n");
    pthread_mutex_lock(&mutex);
    pthread_cond_wait(&cond, &mutex);
    pthread_mutex_unlock(&mutex);
    printf("Processor: Signal received\n");

    printf("Processor: Doing some work...\n");
    atomic_fetch_add(&processed_count, 1);
  }
}

void* Push(void *arg) {
  intptr_t thread_index = (intptr_t)arg;  

  printf("Thread %ld is doing some pseudo work before pushing...\n", thread_index);
  sleep(rand() % 30);

  printf("Thread %ld is pushing...\n", thread_index);
  pthread_mutex_lock(&mutex);
  pthread_cond_signal(&cond);
  pthread_mutex_unlock(&mutex);

  return NULL;
}

int main() {
  pthread_t pusher_threads[THREADS_COUNT];
  pthread_t processor_thread;

  int result = pthread_create(&processor_thread, NULL, (void* (*)(void*))Process, NULL);
  if (result != 0) {
    fprintf(stderr, "Error creating processor thread, result: %d\n", result);
    return 1;
  }

  for (int i = 0; i < THREADS_COUNT; i++) {
    int result = pthread_create(&pusher_threads[i], NULL, Push, (void *)(intptr_t)(i + 1));
    if (result != 0) {
      fprintf(stderr, "Error creating thread %d, result: %d\n", i, result);
      return 1;  
    }
  }

  while(atomic_load(&processed_count) < THREADS_COUNT);

  pthread_mutex_destroy(&mutex);
  pthread_cond_destroy(&cond);

  printf("All threads have finished\n");
  return 0;
}

